//---------------------------------------------------------------------------
// Germination.cpp
//---------------------------------------------------------------------------
#include "Germination.h"
#include "SimManager.h"
#include "TreePopulation.h"
#include "GridBase.h"
#include "ParsingFunctions.h"
#include "ModelMath.h"

///////////////////////////////////////////////////////////////////////////////
//   Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
clGermination::clGermination(clSimManager * p_oSimManager) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    strcpy(m_cNameString, "germination");

    mp_fProportionGerminating = NULL;
    mp_iSeedGridCode = NULL;

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;
  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clGermination::clGermination");
    throw(stcErr);
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
clGermination::~clGermination()
{
  delete[] mp_fProportionGerminating;
  delete[] mp_iSeedGridCode;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   GetData
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clGermination::GetData(DOMDocument * p_oDoc)
{
  try
  {

    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject("treepopulation");
    clTreePopulation * p_oPop = (clTreePopulation *) p_oTemp;
    floatVal * p_fTemp = NULL; //for getting species-specific values
    char cLabel[MAX_GRID_LABEL_SIZE];
    short int iNumSpecies = p_oPop->GetNumberOfSpecies(), i;

    //Declare our arrays
    mp_fProportionGerminating = new float[iNumSpecies];
    mp_iSeedGridCode = new short int[iNumSpecies];

    //Set up our temp array - pre-load with this behavior's species
    p_fTemp = new floatVal[m_iNumBehaviorSpecies];
    for (i = 0; i < m_iNumBehaviorSpecies; i++)
      p_fTemp[i].code = mp_iWhatSpecies[i];

    //Proportion germinating
    FillSpeciesSpecificValue(p_oElement, "ge_proportionGerminating", "ge_pgVal", p_fTemp, m_iNumBehaviorSpecies, p_oPop, true);
    //Now transfer the values to the permanent array
    for (i = 0; i < m_iNumBehaviorSpecies; i++)
      mp_fProportionGerminating[p_fTemp[i].code] = p_fTemp[i].val;

    //Make sure that all values are between 0 and 1
    for (i = 0; i < m_iNumBehaviorSpecies; i++)
    {
      if (mp_fProportionGerminating[mp_iWhatSpecies[i]] < 0 || mp_fProportionGerminating[mp_iWhatSpecies[i]] > 1)
      {
        modelErr stcErr;
        strcpy(stcErr.cFunction, "clGermination::GetData");
        strcpy(stcErr.cMoreInfo, "All values in proportion germinating must be between 0 and 1.");
        stcErr.iErrorCode = BAD_DATA;
        throw(stcErr);
      }
    }

    //Fetch the seed grid
    mp_oSeedGrid = mp_oSimManager->GetGridObject("Dispersed Seeds");
    if (mp_oSeedGrid == NULL)
    {
      modelErr stcErr;
      strcpy(stcErr.cFunction, "clGermination::GetData");
      strcpy(stcErr.cMoreInfo, "Disperse behaviors must be used with germination.");
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      throw(stcErr);
    }

    //Now get the data codes
    for (i = 0; i < m_iNumBehaviorSpecies; i++)
    {
      sprintf(cLabel, "%s%d", "seeds_", mp_iWhatSpecies[i]);
      mp_iSeedGridCode[mp_iWhatSpecies[i]] = mp_oSeedGrid->GetFloatDataCode(cLabel);
      if (-1 == mp_iSeedGridCode[mp_iWhatSpecies[i]])
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy(stcErr.cFunction, "clGermination::SetupGrids");
        strcpy(stcErr.cMoreInfo, "Unexpected grid data return code.");
        throw(stcErr);
      }
    }

    delete[] p_fTemp;
  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clGermination::GetData");
    throw(stcErr);
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   Action
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGermination::Action()
{
  clModelMath *p_oMath;
  float fNumSeeds;
  int iNumXCells = mp_oSeedGrid->GetNumberXCells(),
      iNumYCells = mp_oSeedGrid->GetNumberYCells(),
      iX, iY, i;

  for (iX = 0; iX < iNumXCells; iX++) {
    for (iY = 0; iY < iNumYCells; iY++) {

      //Go through the species
      for (i = 0; i < m_iNumBehaviorSpecies; i++) {

        mp_oSeedGrid->GetValueOfCell(iX, iY, mp_iSeedGridCode[mp_iWhatSpecies[i]], &fNumSeeds);

        fNumSeeds = p_oMath->RandomRound(fNumSeeds * mp_fProportionGerminating[mp_iWhatSpecies[i]]);

        mp_oSeedGrid->SetValueOfCell(iX, iY, mp_iSeedGridCode[mp_iWhatSpecies[i]], fNumSeeds);
      }
    }
  }
}

//---------------------------------------------------------------------------



