//---------------------------------------------------------------------------
// AnisotropicDisperse.cpp
//---------------------------------------------------------------------------
#include "GridBasedSpatialDisperse.h"
#include "SimManager.h"
#include "TreePopulation.h"
#include "ParsingFunctions.h"
#include "GridBase.h"
#include "Plot.h"
#include "Allometry.h"
#include <math.h>
#include <stdlib.h>

///////////////////////////////////////////////////////////////////////////////
//     Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clGridBasedSpatialDisperse::clGridBasedSpatialDisperse( clSimManager * p_oSimManager ) : 
     clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager ), clDisperseBase( p_oSimManager )
     {
       try
       {
         m_iNumFunctions = 2;

         mp_fDispersalX0 = NULL;
         mp_fThetaXb = NULL;
         mp_fStr = NULL;
         mp_fBeta = NULL;
         mp_fFecundity = NULL;
         mp_iIndexes = NULL;
         mp_bIsUsed = NULL;
         mp_iWhatFunction = NULL;

         //Versions
         m_fVersionNumber = 1;
         m_fMinimumVersionNumber = 1;

         //Allowed file types
         m_iNumAllowedTypes = 2;
         mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
         mp_iAllowedFileTypes[0] = parfile;
         mp_iAllowedFileTypes[1] = detailed_output;
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::clGridBasedSpatialDisperse" );
         throw( stcErr );
       }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clGridBasedSpatialDisperse::~clGridBasedSpatialDisperse()
{
  short int i; //loop counter

  if ( mp_bIsUsed )
    for ( i = 0; i < m_iTotalSpecies; i++ )
      delete[] mp_bIsUsed[i];

  delete[] mp_fDispersalX0;
  delete[] mp_fThetaXb;
  delete[] mp_fStr;
  delete[] mp_fBeta;
  delete[] mp_fFecundity;
  delete[] mp_iIndexes;
  delete[] mp_iWhatFunction;
  delete[] mp_bIsUsed;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//     DeclareArrays
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::DeclareArrays()
{
  int i;
  //Declare our data arrays
  mp_fDispersalX0 = new float[m_iNumBehaviorSpecies];
  mp_fThetaXb = new float[m_iNumBehaviorSpecies];
  mp_fBeta = new float[m_iNumBehaviorSpecies];
  mp_fStr = new float[m_iNumBehaviorSpecies];
  mp_fFecundity = new float[m_iNumBehaviorSpecies];
  mp_iWhatFunction = new short int[m_iNumBehaviorSpecies];
  mp_iIndexes = new short int[m_iTotalSpecies];

  //Make the list of indexes
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_iIndexes[mp_iWhatSpecies[i]] = i;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//     DoShellSetup
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::DoShellSetup( xercesc::DOMDocument * p_oDoc )
{
  try
  {

    SetUpBase();

    m_fNumYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();
    mp_oPlot = mp_oSimManager->GetPlotObject();
    m_fSeedCellArea = mp_oSeedGrid->GetLengthXCells() * mp_oSeedGrid->GetLengthYCells();

    DeclareArrays();

    GetParameterFileData( p_oDoc );

    PopulateUsedTable();

    CalculateQueryHeight();

    CalculateFecundity();

    DoChildSetup( p_oDoc );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::GetData" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//     PopulateUsedTable()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::PopulateUsedTable( )
{
  clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
  clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
  short int iNumTypes = p_oPop->GetNumberOfTypes(), i, j; //loop counters

  //Declare the table - it's # species by # types
  mp_bIsUsed = new bool * [m_iTotalSpecies];
  for ( i = 0; i < m_iTotalSpecies; i++ )
  {
    mp_bIsUsed[i] = new bool[iNumTypes];
    for ( j = 0; j < iNumTypes; j++ )
      mp_bIsUsed[i] [j] = false;
  }

  for ( i = 0; i < m_iNumSpeciesTypeCombos; i++ )
  {
    //Make sure that the types are only sapling, adult, or stump
    if ( mp_whatSpeciesTypeCombos[i].iType != clTreePopulation::sapling
         && mp_whatSpeciesTypeCombos[i].iType != clTreePopulation::adult )
         {
           modelErr stcErr;
           strcpy( stcErr.cFunction, "clSpatialDisperse::PopulateUsedTable" );
           strcpy( stcErr.cMoreInfo, "Dispersal may only be applied to saplings and adults." );
           stcErr.iErrorCode = BAD_DATA;
           throw( stcErr );
    }
    mp_bIsUsed[mp_whatSpeciesTypeCombos[i].iSpecies] [mp_whatSpeciesTypeCombos[i].iType] = true;
  }
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
//     GetParameterFileData
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::GetParameterFileData( xercesc::DOMDocument * p_oDoc )
{

  clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
  clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
  DOMElement * p_oElement = p_oDoc->getDocumentElement();
  floatVal * p_fTemp; //for getting species-specific values
  intVal * p_iTemp; //for getting species-specific values
  float fMax; //maximum allowed search distance before an error is thrown
  short int iNumFunctionSpecies, //number of species using a given function
       i; //loop counter

  //Start with which species use which function - influences the rest of
  //the file's reading
  //Set up our temp int array and pre-load with this behavior's species
  p_iTemp = new intVal[m_iNumBehaviorSpecies];
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    p_iTemp[i].code = mp_iWhatSpecies[i];

  p_fTemp = new floatVal[m_iNumBehaviorSpecies];
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    p_fTemp[i].code = mp_iWhatSpecies[i];


  //Minimum DBH for reproduction - remember that this is the static array
  //sized total number of species
  FillSpeciesSpecificValue( p_oElement, "di_minDbhForReproduction", "di_mdfrVal", p_fTemp,
       m_iNumBehaviorSpecies, p_oPop, true );
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fDbhForReproduction[p_fTemp[i].code] = p_fTemp[i].val;

  //What function is always canopy functions
  FillSpeciesSpecificValue( p_oElement, "di_canopyFunction", "di_cfVal", p_iTemp, m_iNumBehaviorSpecies, p_oPop, true );

  //Transfer the values over - if any are not a valid function enum number,
  //throw an error
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    if ( p_iTemp[i].val < 0 || p_iTemp[i].val >= m_iNumFunctions )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::GetData" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Invalid disperse function code ", p_iTemp[i].val );
      throw( stcErr );
    }
    else
    {
      mp_iWhatFunction[mp_iIndexes[mp_iWhatSpecies[i]]] = p_iTemp[i].val;
    }

  //**********************************
  //Weibull parameters
  //**********************************
  //Start by getting a list of species that use weibull - we'll only
  //require parameters for those species
  //First count 'em
  iNumFunctionSpecies = 0;
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    if ( weibull == mp_iWhatFunction[i] )
      iNumFunctionSpecies++;

  //Now declare the float temp array to match and load with these species
  p_fTemp = new floatVal[iNumFunctionSpecies];
  iNumFunctionSpecies = 0;
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    if ( weibull == mp_iWhatFunction[i] )
    {
      p_fTemp[iNumFunctionSpecies].code = mp_iWhatSpecies[i];
      iNumFunctionSpecies++;
    }
  }

  if ( iNumFunctionSpecies > 0 )
  {

    //Dispersal
    FillSpeciesSpecificValue( p_oElement, "di_weibullCanopyDispersal", "di_wcdVal", p_fTemp,
         iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fDispersalX0[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //Theta
    FillSpeciesSpecificValue( p_oElement, "di_weibullCanopyTheta", "di_wctVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fThetaXb[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //Beta
    FillSpeciesSpecificValue( p_oElement, "di_weibullCanopyBeta", "di_wcbVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fBeta[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //STR
    FillSpeciesSpecificValue( p_oElement, "di_weibullCanopySTR", "di_wcsVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fStr[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;
  } //end of if (iNumFunctionSpecies > 0) {

  //**********************************
  //Lognormal parameters
  //**********************************
  //Start by getting a list of species that use lognormal - we'll
  //only require parameters for those species
  //First count 'em
  iNumFunctionSpecies = 0;
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    if ( lognormal == mp_iWhatFunction[i] )
      iNumFunctionSpecies++;

  //Now declare the float temp array to match and load with these species
  delete[] p_fTemp;
  p_fTemp = new floatVal[iNumFunctionSpecies];
  iNumFunctionSpecies = 0;
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    if ( lognormal == mp_iWhatFunction[i] )
    {
      p_fTemp[iNumFunctionSpecies].code = mp_iWhatSpecies[i];
      iNumFunctionSpecies++;
    }
  }

  if ( iNumFunctionSpecies > 0 )
  {

    //X0
    FillSpeciesSpecificValue( p_oElement, "di_lognormalCanopyX0", "di_lcx0Val", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fDispersalX0[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //Xb
    FillSpeciesSpecificValue( p_oElement, "di_lognormalCanopyXb", "di_lcxbVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fThetaXb[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //Beta
    FillSpeciesSpecificValue( p_oElement, "di_lognormalCanopyBeta", "di_lcbVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fBeta[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

    //STR
    FillSpeciesSpecificValue( p_oElement, "di_lognormalCanopySTR", "di_lcsVal", p_fTemp, iNumFunctionSpecies, p_oPop, true );
    //Transfer to permanent array
    for ( i = 0; i < iNumFunctionSpecies; i++ )
      mp_fStr[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;
  } //end of if (iNumFunctionSpecies > 0)

  //Maximum search radius for parents
  //Get the max allowed
  fMax = mp_oPlot->GetXPlotLength() > mp_oPlot->GetYPlotLength() ? mp_oPlot->GetYPlotLength() : mp_oPlot->GetXPlotLength();
  fMax *= 0.5;
  FillSingleValue( p_oElement, "di_maxSearchRadius", & m_fMaxSearchRadius, true );
  //Make sure that this value is not negative or greater than half the shortest
  //plot value
  if ( 0 >= m_fMaxSearchRadius || fMax < m_fMaxSearchRadius )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::GetData" );
    strcpy( stcErr.cMoreInfo, "Max search radius must be between 0 and half the shortest plot length." );
    throw( stcErr );
  }

  delete[] p_iTemp;
  delete[] p_fTemp;
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
//     CalculateFecundity
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::CalculateFecundity()
{
  int i, iSp;
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    iSp = mp_iIndexes[mp_iWhatSpecies[i]];
    mp_fFecundity[iSp] = mp_fStr[iSp] / pow( 30, mp_fBeta[iSp] );

  }
}
//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
// AddSeeds()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::AddSeeds()
{
  try
  {
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
    clPlot * p_oPlot = mp_oSimManager->GetPlotObject();
    clTreeSearch * p_oAllTrees; //search object for getting all trees
    clTree * p_oTree; //for working with a single tree
    char cQuery[75]; //format search strings into this
    float *p_fSeeds = new float[m_iNumBehaviorSpecies], //seed totals for each species
         fDbh, //tree's dbh
         fCellSeeds, //number of seeds existing in a cell
         fCellXLength = mp_oSeedGrid->GetLengthXCells(), fCellYLength = mp_oSeedGrid->GetLengthYCells(),
         fHalfX = fCellXLength * 0.5, fHalfY = fCellYLength * 0.5,
         fHalfEndX, fHalfEndY, //half of the cells in the last rows
         fCellX, fCellY, //coordinates of grid cell into which seeds are being
         //dispersed
         fTreeX, fTreeY; //tree coordinates
    int iNumXCells = mp_oSeedGrid->GetNumberXCells(), iNumYCells = mp_oSeedGrid->GetNumberYCells(), iX, iY,
        iSp, iType, i; //species and type of a given tree

    //The last cells in the row might be short; calculate their half points
    //separately
    fHalfEndX = (p_oPlot->GetXPlotLength() - ((float)(iNumXCells - 1)*fCellXLength)) * 0.5;
    fHalfEndY = (p_oPlot->GetYPlotLength() - ((iNumYCells - 1)*fCellYLength)) * 0.5;

    //Loop through the grid cells in the seed grid
    for ( iX = 0; iX < iNumXCells; iX++ )
    {
      for ( iY = 0; iY < iNumYCells; iY++ )
      {

        //Zero out seed totals array
        for (iSp = 0; iSp < m_iNumBehaviorSpecies; iSp++) {
          p_fSeeds[iSp] = 0;
        }

        //Get the coordinates of the center of this cell
        //If this is the last cell, make sure to correct if it's a short one
        if (iX == iNumXCells - 1) {
          fCellX = ( ( float )iX * fCellXLength ) + fHalfEndX;
        } else {
          fCellX = ( ( float )iX * fCellXLength ) + fHalfX;
        }
        if (iY == iNumYCells - 1) {
          fCellY = ( ( float )iY * fCellYLength ) + fHalfEndY;
        } else {
          fCellY = ( ( float )iY * fCellYLength ) + fHalfY;
        }

        //Construct our distance/height query
        sprintf( cQuery, "%s%f%s%f%s%f%s%f", "distance=", m_fMaxSearchRadius, "FROM x=", fCellX, "y=", fCellY,
             "::height=", m_fQueryHeight );

        //Get the trees that apply
        p_oAllTrees = p_oPop->Find( cQuery );

        //Go through the trees one at a time
        p_oTree = p_oAllTrees->NextTree();
        while ( p_oTree )
        {
          //Get the tree's species and type
          iSp = p_oTree->GetSpecies();
          iType = p_oTree->GetType();

          //Check to see if this species/type combo is used
          if ( mp_bIsUsed[iSp] [iType] )
          {

            //Get the tree's dbh
            p_oTree->GetValue( p_oPop->GetDbhCode( iSp, iType ), & fDbh );

            //If the tree is larger than the minimum size for reproduction,
            //add in its seeds
            if ( fDbh >= mp_fDbhForReproduction[iSp] ) {

              //Get the tree's X and Y
              p_oTree->GetValue(p_oPop->GetXCode(iSp, iType), &fTreeX);
              p_oTree->GetValue(p_oPop->GetYCode(iSp, iType), &fTreeY);

              //Get the fecundity and multiply it by the overridden function
              //value to get the number of seeds
              for (i = 0; i < m_iNumBehaviorSpecies; i++) {
                if (mp_iWhatSpecies[i] == iSp) {
                  p_fSeeds[i] += mp_fFecundity[mp_iIndexes[iSp]] *
                   pow(fDbh, mp_fBeta[mp_iIndexes[iSp]]) *
                   GetFunctionValue(fCellX, fCellY, fTreeX, fTreeY, iSp);
                }
              }
            }
          }

          p_oTree = p_oAllTrees->NextTree();
        } //end of while (p_oTree)

        //RandomRound all seed totals, scale them to the correct number of
        //timesteps and square meters per grid cell, and assign them to the
        //seed grid
        for (iSp = 0; iSp < m_iNumBehaviorSpecies; iSp++) {
          p_fSeeds[iSp] = p_fSeeds[iSp] * m_fNumYearsPerTimestep * m_fSeedCellArea;
          mp_oSeedGrid->GetValueOfCell( iX, iY, mp_iNumSeedsCode[mp_iWhatSpecies[iSp]], &fCellSeeds );
          fCellSeeds += p_fSeeds[iSp];
          mp_oSeedGrid->SetValueOfCell( iX, iY, mp_iNumSeedsCode[mp_iWhatSpecies[iSp]], fCellSeeds);
        }
      }
    }

    delete[] p_fSeeds;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::AddSeeds" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
// CalculateQueryHeight()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clGridBasedSpatialDisperse::CalculateQueryHeight() {
  try {
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
    clAllometry *p_oAllom = p_oPop->GetAllometryObject();
    float fHeight;
    int i, iSp;

    m_fQueryHeight = 10000;

    for (i = 0; i < m_iNumBehaviorSpecies; i++) {
      iSp = mp_iIndexes[mp_iWhatSpecies[i]];
      if (p_oPop->GetMinAdultDBH(iSp) > mp_fDbhForReproduction[iSp]) {
        //This is a sapling
        fHeight = p_oAllom->CalcSaplingHeight(mp_fDbhForReproduction[iSp], mp_iWhatSpecies[i]);
      } else {
        //An adult
        fHeight = p_oAllom->CalcAdultHeight(mp_fDbhForReproduction[iSp], mp_iWhatSpecies[i]);
      }
      if (m_fQueryHeight > fHeight) {
        m_fQueryHeight = fHeight;
      }
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clGridBasedSpatialDisperse::CalculateQueryHeight" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------


