//---------------------------------------------------------------------------
// NeighborhoodSeedPredation.cpp
//---------------------------------------------------------------------------
#include "NeighborhoodSeedPredation.h"
#include "SimManager.h"
#include "TreePopulation.h"
#include "GridBase.h"
#include "ParsingFunctions.h"
#include "ModelMath.h"
#include "Plot.h"
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// May 21, 2007 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
clNeighborhoodSeedPredation::clNeighborhoodSeedPredation( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    strcpy( m_cNameString, "Neighborhood Seed Predation" );
         
    mp_fNonMastingA = NULL;
    mp_fNonMastingB = NULL;
    mp_fNonMastingC = NULL;
    mp_fMastingA = NULL;
    mp_fMastingB = NULL;
    mp_fMastingC = NULL;
    mp_fMastingLoadingFactor = NULL;
    mp_fNonMastingLoadingFactor = NULL;
    mp_iSeedGridCode = NULL;
    mp_iIndexes = NULL;

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::clNeighborhoodSeedPredation" );
    throw( stcErr );
  }
}

///////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// May 21, 2007 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
clNeighborhoodSeedPredation::~clNeighborhoodSeedPredation()
{
  delete[] mp_fNonMastingA;
  delete[] mp_fNonMastingB;
  delete[] mp_fNonMastingC;
  delete[] mp_fMastingA;
  delete[] mp_fMastingB;
  delete[] mp_fMastingC;
  delete[] mp_fMastingLoadingFactor;
  delete[] mp_fNonMastingLoadingFactor;
  delete[] mp_iSeedGridCode;
  delete[] mp_iIndexes;
}

///////////////////////////////////////////////////////////////////////////////
// GetData
// Edit history:
// -------------
// May 21, 2007 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clNeighborhoodSeedPredation::GetData( DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    clPlot *p_oPlot = mp_oSimManager->GetPlotObject();
    floatVal * p_fTemp = NULL; //for getting species-specific values
    char cLabel[MAX_GRID_LABEL_SIZE];
    short int iNumSpecies = p_oPop->GetNumberOfSpecies(), i;
    
    //Declare the indexes array
    mp_iIndexes = new short int[iNumSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_iIndexes[mp_iWhatSpecies[i]] = i;

    //Declare our arrays
    mp_fNonMastingA = new float[m_iNumBehaviorSpecies];
    mp_fNonMastingB = new float[m_iNumBehaviorSpecies];
    mp_fNonMastingC = new float[m_iNumBehaviorSpecies];
    mp_fMastingA = new float[m_iNumBehaviorSpecies];
    mp_fMastingB = new float[m_iNumBehaviorSpecies];
    mp_fMastingC = new float[m_iNumBehaviorSpecies];
    mp_iSeedGridCode = new short int[m_iNumBehaviorSpecies];
    mp_fMastingLoadingFactor = new float[iNumSpecies];
    mp_fNonMastingLoadingFactor = new float[iNumSpecies];

    //Set up our temp array - pre-load with this behavior's species
    p_fTemp = new floatVal[m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      p_fTemp[i].code = mp_iWhatSpecies[i];

    //Masting "a"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredMastingA", 
                              "pr_npmaVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fMastingA[i] = p_fTemp[i].val;
      
    //Masting "b"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredMastingB", 
                              "pr_npmbVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fMastingB[i] = p_fTemp[i].val;
    
    //Masting "c"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredMastingC", 
                              "pr_npmcVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fMastingC[i] = p_fTemp[i].val;
      
    //Non-masting "a"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredNonMastingA", 
                              "pr_npnmaVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fNonMastingA[i] = p_fTemp[i].val;
      
    //Non-masting "b"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredNonMastingB", 
                              "pr_npnmbVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fNonMastingB[i] = p_fTemp[i].val;
      
    //Non-masting "c"
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredNonMastingC", 
                              "pr_npnmcVal", 
                              p_fTemp, 
                              m_iNumBehaviorSpecies, 
                              p_oPop, 
                              true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fNonMastingC[i] = p_fTemp[i].val;

    //Masting loading factor - required for all species
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredMastingPCALoadFactor", 
                              "pr_npmplfVal",
                              mp_fMastingLoadingFactor,
                              p_oPop,
                              true); 
                              
    //Nonmasting loading factor - required for all species
    FillSpeciesSpecificValue( p_oElement, 
                              "pr_neighPredNonMastingPCALoadFactor", 
                              "pr_npnmplfVal",
                              mp_fNonMastingLoadingFactor,
                              p_oPop,
                              true);
                              
    //Density of seeds for masting
    FillSingleValue (p_oElement, 
                     "pr_neighPredMastingDensity",
                     &m_fMastingThreshold,
                     true);
    if ( m_fMastingThreshold < 0 )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
      strcpy( stcErr.cMoreInfo, "Masting seed density threshold cannot be negative." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }                 
    //Transform from annual to timestep value
    m_fMastingThreshold *= mp_oSimManager->GetNumberOfYearsPerTimestep();
    
    //Min neighbor DBH
    FillSingleValue (p_oElement, 
                     "pr_neighPredMinNeighDBH",
                     &m_fMinNeighDBH,
                     true);
    if ( m_fMinNeighDBH < 0 )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
      strcpy( stcErr.cMoreInfo, "Minimum neighbor DBH cannot be negative." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }
    
    //Neighborhood search radius
    FillSingleValue (p_oElement, 
                     "pr_neighPredRadius",
                     &m_fRadius,
                     true);
    if ( m_fRadius < 0 )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
      strcpy( stcErr.cMoreInfo, "Neighborhood radius cannot be negative." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }                         

    //Fetch the seed grid
    mp_oSeedGrid = mp_oSimManager->GetGridObject( "Dispersed Seeds" );
    if ( mp_oSeedGrid == NULL )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
      strcpy( stcErr.cMoreInfo, "Disperse behaviors must be used with neighborhood seed predation." );
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      throw( stcErr );
    }

    //Now get the data codes
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      sprintf( cLabel, "%s%d", "seeds_", mp_iWhatSpecies[i] );
      mp_iSeedGridCode[i] = mp_oSeedGrid->GetFloatDataCode( cLabel );
      if ( -1 == mp_iSeedGridCode[i] )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
        strcpy( stcErr.cMoreInfo, "Unexpected grid data return code." );
        throw( stcErr );
      }
    }

    //Calculate the plot areas
    m_fPlotArea = p_oPlot->GetXPlotLength() * p_oPlot->GetYPlotLength();

    delete[] p_fTemp;
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::GetData" );
    throw( stcErr );
  }
}

///////////////////////////////////////////////////////////////////////////////
// Action
// Edit history:
// -------------
// May 21, 2007 - Created (LEM)
///////////////////////////////////////////////////////////////////////////////
void clNeighborhoodSeedPredation::Action()
{
  try {	
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreeSearch * p_oTrees;             //saplings and adults
    clTree * p_oTree;                    //single tree to count
    char cQuery[75];                     //format search string into this
    float *p_fA, *p_fB, *p_fC, *p_fLoad, //stand-in for parameter arrays
          *p_fBasalArea,                 //basal area per species
          fNumSeeds = 0, fNewSeeds,
          fTotalBA,
          fPCA, fDbh,
          fX, fY;
    int iNumXCells = mp_oSeedGrid->GetNumberXCells(), 
        iNumYCells = mp_oSeedGrid->GetNumberYCells(), 
        iNumSpecies = p_oPop->GetNumberOfSpecies(),
        iX, iY, iSp, iTp;
        
    p_fBasalArea = new float[iNumSpecies];
    
    //Are we masting? Find the seed density
    for ( iX = 0; iX < iNumXCells; iX++ ) {
      for ( iY = 0; iY < iNumYCells; iY++ ) {
        for (iSp = 0; iSp < m_iNumBehaviorSpecies; iSp++) {
          mp_oSeedGrid->GetValueOfCell(iX, iY, mp_iSeedGridCode[iSp], &fNewSeeds);
          fNumSeeds += fNewSeeds;    
        }
      }
    }
    
    //Set the parameters appropriate to the situation
    if (fNumSeeds / m_fPlotArea <= m_fMastingThreshold) {
      //Non-masting
      p_fA = mp_fNonMastingA;
      p_fB = mp_fNonMastingB;
      p_fC = mp_fNonMastingC;
      p_fLoad = mp_fNonMastingLoadingFactor;
    } else {
      //Masting
      p_fA = mp_fMastingA;
      p_fB = mp_fMastingB;
      p_fC = mp_fMastingC;
      p_fLoad = mp_fMastingLoadingFactor;
    }
    	

    for ( iX = 0; iX < iNumXCells; iX++ )
    {
      for ( iY = 0; iY < iNumYCells; iY++ )
      {
      
        //Get all neighbors within the specified radius and calculate 
        //basal area
        for (iSp = 0; iSp < iNumSpecies; iSp++) p_fBasalArea[iSp] = 0;
        mp_oSeedGrid->GetPointOfCell( iX, iY, & fX, & fY );
        sprintf( cQuery, "%s%f%s%f%s%f%s%f", "distance=", m_fRadius, "FROM x=", 
           fX, "y=", fY, "::height=", 0.0 );
        p_oTrees = p_oPop->Find(cQuery);
        p_oTree = p_oTrees->NextTree();
        while (NULL != p_oTree) {

          //Get the tree's DBH and make sure it's eligible
          iSp = p_oTree->GetSpecies();
          iTp = p_oTree->GetType();
          if (iTp == clTreePopulation::sapling || 
              iTp == clTreePopulation::adult) {
            
            p_oTree->GetValue(p_oPop->GetDbhCode(iSp, iTp), &fDbh);

            if (fDbh >= m_fMinNeighDBH) {
        
              //Get tree's "basal area" - since it's relative BA, we really
              //only need to square the DBH
              p_fBasalArea[iSp] += pow(fDbh, 2);
            }
          }
          p_oTree = p_oTrees->NextTree();
        }
        
        //Get relative BA for each species
        fTotalBA = 0; fPCA = 0;
        for (iSp = 0; iSp < iNumSpecies; iSp++) fTotalBA += p_fBasalArea[iSp];
        if (fTotalBA > 0) {
          for (iSp = 0; iSp < iNumSpecies; iSp++) 
            fPCA += (p_fBasalArea[iSp] / fTotalBA) * p_fLoad[iSp];
        } else fPCA = 0;
          
        //Go through the species and reduce seeds appropriately
        for ( iSp = 0; iSp < m_iNumBehaviorSpecies; iSp++ )
        {
          mp_oSeedGrid->GetValueOfCell( iX, iY, mp_iSeedGridCode[iSp], & fNumSeeds );

          //Reduce the number of seeds appropriately
          fNewSeeds = p_fA[iSp] * pow(p_fB[iSp], pow(fPCA - p_fC[iSp], 2));
          fNewSeeds = clModelMath::RandomRound(fNumSeeds * (1 - fNewSeeds));
          
          if ( fNumSeeds < 0 ) fNumSeeds = 0;
          if ( fNewSeeds > fNumSeeds) fNewSeeds = fNumSeeds;

          mp_oSeedGrid->SetValueOfCell( iX, iY, mp_iSeedGridCode[iSp], fNewSeeds );
        }   
      }
    }  

    delete[] p_fBasalArea;
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clNeighborhoodSeedPredation::Action" );
    throw( stcErr );
  }
}
